"""Models module."""
from .base import FeedForwardPredictor
from .parser import DependencyRelationModel
from .embeddings import CharacterBasedWordEmbeddings
from .encoder import ComboEncoder
from .lemma import LemmatizerModel
from .model import SemanticMultitaskModel
from .morpho import MorphologicalFeatures
